/* This file is part of the KDE project
   Copyright (C) 2003-2004 Nadeem Hasan <nhasan@kde.org>
   Copyright (C) 2004 Aaron J. Seigo <aseigo@kde.org>
   Copyright (C) 2004 Georges A.K. <georgesak@gmail.com>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "hidebutton.h"

#include <qpainter.h>

#include <kapplication.h>
#include <kcursor.h>
#include <kglobalsettings.h>
#include <kiconeffect.h>
#include <kicontheme.h>
#include <kipc.h>
#include <kstandarddirs.h>
#include <kiconloader.h>
#include <qdir.h>

HideButton::HideButton( QWidget *parent, const char *name )
	: QButton( parent, name ), mHighlighted( false ), mDirection(0)
{
	setBackgroundOrigin( AncestorOrigin );

	connect( kapp, SIGNAL( settingsChanged( int ) ),
	SLOT( slotSettingsChanged( int ) ) );
	connect( kapp, SIGNAL( iconChanged( int ) ),
	SLOT( slotIconChanged( int ) ) );

	kapp->addKipcEventMask( KIPC::SettingsChanged );
	kapp->addKipcEventMask( KIPC::IconChanged );

	slotSettingsChanged( KApplication::SETTINGS_MOUSE );
}

void HideButton::drawButton( QPainter *p )
{
	drawButtonLabel( p );
}

void HideButton::drawButtonLabel( QPainter *p )
{
	if (pixmap())
	{
		QPixmap pix = mHighlighted? mActiveIcons[mDirection] : mNormalIcons[mDirection];

		if ( isOn() || isDown() )
		{
			p->translate( 2, 2 );
		}

		QSize point = margin()/2;
		QPoint origin(point.width(), point.height());

		if (pix.height() < (height() - margin().height()))
		{
			origin.setY(origin.y() + ((height() - pix.height()) / 2));
		}

		if (pix.width() < (width() - margin().width()))
		{
			origin.setX(origin.x() + ((width() - pix.width()) / 2));
		}

		p->drawPixmap(origin, pix);
	}
}

void HideButton::setPixmap( const QPixmap &pix )
{
	QButton::setPixmap( pix );
	generateIcons();
	resize(pix.width()+margin().width()/2, pix.height()+margin().height()/2);
}

void HideButton::generateIcons()
{
	if ( !pixmap() )
		return;

	QImage image = pixmap()->convertToImage();
	KIconEffect effect;

	mNormalIcons[0] = effect.apply( image, KIcon::Panel, KIcon::DefaultState );
	mActiveIcons[0] = effect.apply( image, KIcon::Panel, KIcon::ActiveState );
	mNormalIcons[1] = mNormalIcons[0].xForm(QWMatrix().rotate(90));
	mActiveIcons[1] = mActiveIcons[0].xForm(QWMatrix().rotate(90));
	mNormalIcons[2] = mNormalIcons[1].xForm(QWMatrix().rotate(90));
	mActiveIcons[2] = mActiveIcons[1].xForm(QWMatrix().rotate(90));
	mNormalIcons[3] = mNormalIcons[2].xForm(QWMatrix().rotate(90));
	mActiveIcons[3] = mActiveIcons[2].xForm(QWMatrix().rotate(90));
}

void HideButton::slotSettingsChanged( int category )
{
	if ( category != KApplication::SETTINGS_MOUSE ) return;

	bool changeCursor = KGlobalSettings::changeCursorOverIcon();

	if ( changeCursor )
		setCursor( KCursor::handCursor() );
	else
		unsetCursor();
}

void HideButton::slotIconChanged( int group )
{
	if ( group != KIcon::Panel )
		return;

	generateIcons();
	repaint( false );
}

void HideButton::enterEvent( QEvent *e )
{
	mHighlighted = true;

	repaint( false );
	QButton::enterEvent( e );
}

void HideButton::leaveEvent( QEvent *e )
{
	mHighlighted = false;

	repaint( false );
	QButton::enterEvent( e );
}

void HideButton::resizeEvent( QResizeEvent * )
{
	generateIcons();
}

void HideButton::mousePressEvent(QMouseEvent* e)
{
	if (e->button() == QMouseEvent::RightButton) {
		emit rightClick();
	}

	QButton::mousePressEvent(e);
}

void HideButton::showLeftArrow()
{
	mDirection = 0;
	repaint( false );
}

void HideButton::showRightArrow()
{
	mDirection = 2;
	repaint( false );
}

void HideButton::showUpArrow()
{
	mDirection = 3;
	repaint( false );
}

void HideButton::showDownArrow()
{
	mDirection = 1;
	repaint( false );
}

void HideButton::loadIcon(QString pIconName)
{
	if (pIconName == "KDE Default 1") {
		setPixmap(kapp->iconLoader()->loadIcon("1leftarrow", KIcon::Small));	
		return;
	}

	if (pIconName == "KDE Default 2") {
		setPixmap(kapp->iconLoader()->loadIcon("2leftarrow", KIcon::Small));	
		return;
	}
	setPixmap(QDir::homeDirPath() + QString("/.kde/share/apps/kicker/applet/systemtray2/icons/") + pIconName);	
}

#include "hidebutton.moc"
