/*****************************************************************

Copyright (c) 2000-2001 Matthias Ettrich <ettrich@kde.org>
              2000-2001 Matthias Elter   <elter@kde.org>
              2001      Carsten Pfeiffer <pfeiffer@kde.org>
              2001      Martijn Klingens <mklingens@yahoo.com>
              2004      Georges A.Khalil <georgesak@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#include <qcursor.h>

#include <kwin.h>
#include <krun.h>
#include <kglobal.h>
#include <kapplication.h>
#include <klocale.h>
#include <kwinmodule.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kstandarddirs.h>
#include <kconfigdialog.h>
#include <kaboutdialog.h>
#include <kapplication.h>

#include <qimage.h>
#include <qtimer.h>

#include "hidebutton.h"
#include "configdlg.h"
#include "systemtrayapplet2.h"

#include <X11/Xlib.h>

extern "C"
{
	KPanelApplet* init( QWidget *parent, const QString& configFile)
	{
		KGlobal::locale()->insertCatalogue("systemtray2");
		return new SystemTrayApplet2(configFile, KPanelApplet::Normal,
							/*KPanelApplet::About | */KPanelApplet::Preferences,
							parent, "systemtray2");
	}
}


SystemTrayApplet2::SystemTrayApplet2(const QString& configFile, Type type, int actions,
								QWidget *parent, const char *name)
: KPanelApplet(configFile, type, actions, parent, name),
	showFrame(false), mCollapsed(true)
{
	mHideButton = new HideButton(this);
	connect(mHideButton, SIGNAL(pressed()), this, SLOT(hideButtonClicked()));
	connect(mHideButton, SIGNAL(rightClick()), this, SLOT(preferences()));

	mAutoHideTimer = new QTimer(this);
	connect(mAutoHideTimer, SIGNAL(timeout()), this, SLOT(hideButtonClicked()));

	mSmoothScrollTimer = new QTimer(this);
	connect(mSmoothScrollTimer, SIGNAL(timeout()), this, SLOT(slotSmoothScrollTimer()));
	
	config()->setGroup("General");
	KConfig *globalConf = KGlobal::config();
	globalConf->setGroup("General");

	if (config()->readBoolEntry( "ShowPanelFrame", false ))
	{
		setFrameStyle(Panel | Sunken);
	}
	else if (globalConf->readBoolEntry("FadeOutAppletHandles", false) ||
			globalConf->readBoolEntry("HideAppletHandles", false))
	{
		// only do the "show on hover" trick with the frame when we DON'T have handles
		showFrame = true;
	}

	m_Wins.setAutoDelete(true);
	setBackgroundOrigin(AncestorOrigin);

	kwin_module = new KWinModule(this);

	// register existing tray windows
	const QValueList<WId> systemTrayWindows = kwin_module->systemTrayWindows();
	bool existing = false;
	for (QValueList<WId>::ConstIterator it = systemTrayWindows.begin(); it!=systemTrayWindows.end(); ++it ) {
		embedWindow( *it, true );
		existing = true;
	}
	if (existing)
		layoutTray();

	// the KWinModule notifies us when tray windows are added or removed
	connect( kwin_module, SIGNAL( systemTrayWindowAdded(WId) ),
			this, SLOT( systemTrayWindowAdded(WId) ) );
	connect( kwin_module, SIGNAL( systemTrayWindowRemoved(WId) ),
			this, SLOT( updateTrayWindows() ) );

	QCString screenstr;
	screenstr.setNum(qt_xscreen());
	QCString trayatom = "_NET_SYSTEM_TRAY_S" + screenstr;

	Display *display = qt_xdisplay();

	net_system_tray_selection = XInternAtom( display, trayatom, FALSE );
	net_system_tray_opcode = XInternAtom( display, "_NET_SYSTEM_TRAY_OPCODE", FALSE );

	// Acquire system tray
	XSetSelectionOwner(display,
					net_system_tray_selection,
					winId(),
					CurrentTime);

	WId root = qt_xrootwin();

	if (XGetSelectionOwner (display, net_system_tray_selection) == winId())
	{
		XClientMessageEvent xev;

		xev.type = ClientMessage;
		xev.window = root;

		xev.message_type = XInternAtom (display, "MANAGER", False);
		xev.format = 32;
		xev.data.l[0] = CurrentTime;
		xev.data.l[1] = net_system_tray_selection;
		xev.data.l[2] = winId();
		xev.data.l[3] = 0;        /* manager specific data */
		xev.data.l[4] = 0;        /* manager specific data */

		XSendEvent (display, root, False, StructureNotifyMask, (XEvent *)&xev);
	}

	resize(0,0);

	slotLoadSettings();
}

SystemTrayApplet2::~SystemTrayApplet2()
{
	m_Wins.clear();
	KGlobal::locale()->removeCatalogue("ksystemtrayapplet");
}

bool SystemTrayApplet2::x11Event( XEvent *e )
{
#define SYSTEM_TRAY_REQUEST_DOCK    0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2
	if ( e->type == ClientMessage ) {
		if ( e->xclient.message_type == net_system_tray_opcode &&
			e->xclient.data.l[1] == SYSTEM_TRAY_REQUEST_DOCK) {
			for (QXEmbed* emb = m_Wins.first(); emb != 0L; emb = m_Wins.next())
				if( emb->embeddedWinId() == (WId)e->xclient.data.l[2] ) // we already manage it
					return true;
			embedWindow( e->xclient.data.l[2], false );
			layoutTray();
			emit updateLayout();
			return true;
		}
	}
	return KPanelApplet::x11Event( e ) ;
}

void SystemTrayApplet2::enterEvent( QEvent * )
{
	/*
	* NOTE: This doesn't actually work as one would expect.  Since system tray
	* icons are actually separate windows (or something like that) enter and
	* leave events are triggered when the mouse goes over the icons.
	*
	* This makes it possible to "skip" over the window (i.e. leave the system
	* tray while a popup menu from one of the systray icons is active) and no
	* leave event will be sent, thus leaving the border shown.
	*/

	if ( showFrame && ( frameStyle() == NoFrame ) )
	{
		setFrameStyle( Panel|Sunken );
		repaint();
	}

	if (mAutoHideTimer->isActive()) {
		mAutoHideTimer->stop();
	}
}

void SystemTrayApplet2::leaveEvent( QEvent * )
{
	if ( showFrame && !geometry().contains( mapFromGlobal( QCursor::pos() ) ) )
	{
		setFrameStyle( NoFrame );
		repaint();
	}

	if (Config::iconHiding() && Config::autoCollapse() &&
		!mCollapsed && !geometry().contains(mapFromGlobal(QCursor::pos()))) {
		mAutoHideTimer->start(2000, true);
	}
}

void SystemTrayApplet2::systemTrayWindowAdded( WId w )
{
	for (QXEmbed* emb = m_Wins.first(); emb != 0L; emb = m_Wins.next())
		if( emb->embeddedWinId() == w ) // we already manage it
			return;
	embedWindow( w, true );
	layoutTray();
	emit updateLayout();
}

void SystemTrayApplet2::embedWindow( WId w, bool kde_tray )
{
	TrayEmbed* emb = new TrayEmbed(kde_tray,this);
	emb->setAutoDelete(false);
	emb->setBackgroundOrigin(AncestorOrigin);
	emb->setBackgroundMode(X11ParentRelative);

	if( kde_tray )
	{
		static Atom hack_atom = XInternAtom( qt_xdisplay(), "_KDE_SYSTEM_TRAY_EMBEDDING", False );
		XChangeProperty( qt_xdisplay(), w, hack_atom, hack_atom, 32, PropModeReplace, NULL, 0 );
		emb->embed(w);
		XDeleteProperty( qt_xdisplay(), w, hack_atom );
	}
	else
		emb->embed(w);
		
	if( emb->embeddedWinId() == 0 ) // error embedding
	{
		delete emb;
		return;
	}

	connect(emb, SIGNAL(embeddedWindowDestroyed()), SLOT(updateTrayWindows()));
	emb->resize(24, 24);
	emb->show();
	if (Config::hiddenList().contains(KWin::WindowInfo(w, 0, 0).visibleName())) {
		emb->setCollapsed(true);
	}

	if (Config::priorityList().contains(KWin::WindowInfo(w, 0, 0).visibleName())) {
		int i = Config::priorityList().findIndex(KWin::WindowInfo(w, 0, 0).visibleName());
		emb->setPriority(Config::priorityList()[i+1].toInt());
	}

	m_Wins.inSort(emb);
}

void SystemTrayApplet2::updateTrayWindows()
{
	int i = 0;
	TrayEmbed* emb = m_Wins.first();
	while ((emb = m_Wins.current()) != 0L) {
		WId wid = emb->embeddedWinId();
		if ((wid == 0)
			|| ( emb->kdeTray() && !kwin_module->systemTrayWindows().contains(wid)))
		{
//            emb->setMapAfterRelease( false ); // don't map, they went away on their own
			m_Wins.remove(i);
		}
		else {
			m_Wins.next();
			i++;
		}
	}
	layoutTray();
	emit updateLayout();
}

int SystemTrayApplet2::maxIconWidth() const
{
	int largest = 24;
	for (TrayEmbedList::const_iterator emb = m_Wins.begin(); emb != m_Wins.end(); ++emb)
	{
		if (*emb == 0)
		{
			continue;
		}

		int buttonWidth = (*emb)->sizeHint().width();
		if (buttonWidth > largest)
		{
			largest = buttonWidth;
		}
	}

	return largest;
}

int SystemTrayApplet2::maxIconHeight() const
{
	int largest = 24;
	for (TrayEmbedList::const_iterator emb = m_Wins.begin(); emb != m_Wins.end(); ++emb)
	{
		if (*emb == 0)
		{
			continue;
		}

		int buttonHeight = (*emb)->sizeHint().height();
		if (buttonHeight > largest)
		{
			largest = buttonHeight;
		}
	}

	return largest;
}

int SystemTrayApplet2::widthForHeight(int h) const
{
	const int INTERVAL = 10;
	int iconWidth = maxIconWidth(), iconHeight = maxIconHeight();

	if (h < iconHeight)
	{
		// avoid div by 0 later
		h = iconHeight;
	}

	int count = 0;
	if (Config::iconHiding() && mCollapsed) {
		TrayEmbedList Temp = m_Wins;
		for (TrayEmbed* emb = Temp.first(); emb != 0L; emb = Temp.next()) {
			if (emb->isCollapsed() == false) {
				count++;
			}
		}
	} else {
		count = m_Wins.count();
 	}

	int ret;
	if (Config::iconHiding()) {
		ret = ( ( ( count - 1 ) / ( h / iconHeight ) ) + 1 ) * iconWidth + 4 
			+ mHideButton->width();
	} else {
		ret = ( ( ( count - 1 ) / ( h / iconHeight ) ) + 1 ) * iconWidth + 4;
	}
	if (ret < iconWidth + 4)
	{
		ret = 0;
	}

 	if (Config::smoothScrolling()) {
		if (width() > ret + Config::smoothScrollingSpeed()) {
			ret = width() - Config::smoothScrollingSpeed();
			mSmoothScrollTimer->start(INTERVAL, true);
		} else if (width() < ret - Config::smoothScrollingSpeed()) {
			ret = width() + Config::smoothScrollingSpeed();
			mSmoothScrollTimer->start(INTERVAL, true);
		}
	}

	return ret;
}

int SystemTrayApplet2::heightForWidth(int w) const
{
	const int INTERVAL = 15;
	int iconWidth = maxIconWidth(), iconHeight = maxIconHeight();

	if (w < iconWidth)
	{
		// avoid div by 0 later
		w = iconWidth;
	}

	int count = 0;
	if (Config::iconHiding() && mCollapsed) {
		TrayEmbedList Temp = m_Wins;
		for (TrayEmbed* emb = Temp.first(); emb != 0L; emb = Temp.next()) {
			if (emb->isCollapsed() == false) {
				count++;
			}
		}
	} else {
		count = m_Wins.count();
	}

	int ret;
	if (Config::iconHiding()) {
		ret = ( ( ( count - 1 ) / ( w / iconWidth ) ) + 1 ) * iconHeight + 4 
			+ mHideButton->height();
	} else {
		ret = ( ( ( count - 1 ) / ( w / iconWidth ) ) + 1 ) * iconHeight + 4;
	}

	if (ret < iconHeight + 4)
	{
		ret = 0;
	}

	if (Config::smoothScrolling()) {
		if (height() > ret + Config::smoothScrollingSpeed()) {
			ret = height() - Config::smoothScrollingSpeed();
			mSmoothScrollTimer->start(INTERVAL, true);
	
		} else if (height() < ret - Config::smoothScrollingSpeed()) {
			ret = height() + Config::smoothScrollingSpeed();
			mSmoothScrollTimer->start(INTERVAL, true);
		}
	}

	return ret;
}

void SystemTrayApplet2::resizeEvent( QResizeEvent* )
{
	layoutTray();
}

void SystemTrayApplet2::layoutTray()
{
	if (m_Wins.count() == 0)
		return;

	/* buttonHeightWidth = buttonHeight or buttonWidth in pixels (depends on orientation())
	* nbrOfLines = number of rows or cols (depends on orientation())
	* spacing = size of spacing in pixels between lines (rows or cols) 
	* line = what line to draw an icon in */
	int i = 0, line, spacing, nbrOfLines, heightWidth; 

	int iconWidth = maxIconWidth(), iconHeight = maxIconHeight();

	TrayEmbed* emb;
	int col = 0; // col = column och row, depends on orientation(), the opposit direction of line

	if (orientation() == Vertical) {
		heightWidth = width();
		heightWidth = heightWidth < iconWidth ? iconWidth : heightWidth; // to avoid nbrOfLines=0
		nbrOfLines = heightWidth / iconWidth;
		spacing = ( heightWidth - iconWidth*nbrOfLines ) / ( nbrOfLines + 1 );
		for (emb = m_Wins.first(); emb != 0L; emb = m_Wins.next()) {
			if (Config::iconHiding() && mCollapsed && emb->isCollapsed()) {
				emb->move( 0, -maxIconHeight());
				emb->hide();
				continue;
			}
			emb->show();
			line = i % nbrOfLines;
			if (Config::iconHiding() && !Config::buttonInvertPos()) {
				emb->move( spacing*(line+1) + line*iconWidth, 2 + col*iconHeight 
					+ mHideButton->height());
			} else {
				emb->move( spacing*(line+1) + line*iconWidth, 2 + col*iconHeight);
			}
			if ( line + 1 == nbrOfLines )
				col++;
			i++;
		}

		if (Config::iconHiding()) {
			if (!Config::buttonInvertPos()) {
				mHideButton->move((width()-mHideButton->width())/2, 0);
			} else {
				mHideButton->move((width()-mHideButton->width())/2, height()-mHideButton->height());
			}
			mHideButton->show();
			if (mCollapsed ^ Config::buttonInvertPos()) {
				mHideButton->showDownArrow();
			} else {
				mHideButton->showUpArrow();
			}
		} else {
			mHideButton->hide();
		}
	}
	else {
		heightWidth = height();
		heightWidth = heightWidth < iconHeight ? iconHeight : heightWidth; // to avoid nbrOfLines=0
		nbrOfLines = heightWidth / iconHeight;
		spacing = ( heightWidth - iconHeight*nbrOfLines ) / ( nbrOfLines + 1 );
		for (emb = m_Wins.first(); emb != 0L; emb = m_Wins.next()) {
			if (Config::iconHiding() && mCollapsed && emb->isCollapsed()) {
				emb->move( -maxIconWidth(), 0 );
				emb->hide();
				continue;
			}
			emb->show();
			line = i % nbrOfLines; 
			if (Config::iconHiding() && !Config::buttonInvertPos()) {
				emb->move( 2 + col*iconWidth + mHideButton->width(), 
					spacing*(line+1) + line*iconHeight );
			} else {
				emb->move( 2 + col*iconWidth, spacing*(line+1) + line*iconHeight );
			}
			if ( line + 1 == nbrOfLines )
				col++;
			i++;
		}

		if (Config::iconHiding()) {
			if (!Config::buttonInvertPos()) {
				mHideButton->move(0, (height()-mHideButton->height())/2);
			} else {
				mHideButton->move(width()-mHideButton->width(), (height()-mHideButton->height())/2);
			}
			mHideButton->show();
			if (mCollapsed ^ Config::buttonInvertPos()) {
				mHideButton->showLeftArrow();
			} else {
				mHideButton->showRightArrow();
			}
		} else {
			mHideButton->hide();
		}
	}

	updateGeometry();
}

void SystemTrayApplet2::paletteChange(const QPalette & /* oldPalette */)
{
	for (QXEmbed* emb = m_Wins.first(); emb != 0L; emb = m_Wins.next()) {
		emb->hide();
		emb->show();
	}
}

void SystemTrayApplet2::about()
{
// 	KAboutDialog *lAboutDlg = new KAboutDialog(this);
// 	lAboutDlg->setAuthor("Georges A.K.", "georgesak@gmail.com", "", "");
// 	lAboutDlg->setMaintainer("Georges A.K.", "georgesak@gmail.com", "", "");
// 	lAboutDlg->setVersion("0.4");
// 	lAboutDlg->show();
// //	delete lAboutDlg;
}

void SystemTrayApplet2::preferences()
{
	if (configDlg::showDialog("Preferences")) {
		return;
	}

	configDlg* lConfigDialog = new configDlg(this, "Preferences", &m_Wins);

	connect(lConfigDialog, SIGNAL(settingsChanged()), this, SLOT(slotLoadSettings()));

	lConfigDialog->setMinimumSize(400, 400);
	lConfigDialog->show();
}

void SystemTrayApplet2::slotLoadSettings()
{
	mHideButton->loadIcon(Config::buttonIcon());

	for (TrayEmbed* lEmb=m_Wins.first();lEmb!=0L;lEmb=m_Wins.next()) {
		QString lWName = KWin::WindowInfo(lEmb->embeddedWinId(), 0, 0).visibleName();
		if (Config::hiddenList().contains(lWName)) {
			lEmb->setCollapsed(true);
		} else {
			lEmb->setCollapsed(false);
		}

		if (Config::priorityList().contains(lWName)) {
			int i = Config::priorityList().findIndex(lWName);
			lEmb->setPriority(Config::priorityList()[i+1].toInt());
		} else {
			lEmb->setPriority(0);
		}
	}

	m_Wins.sort();

	layoutTray();
	emit updateLayout();
}

void SystemTrayApplet2::hideButtonClicked()
{
	mCollapsed = !mCollapsed;
	layoutTray();
	emit updateLayout();
}

void SystemTrayApplet2::slotSmoothScrollTimer()
{
	emit updateLayout();
}
